from prism_ai.api_resources.api_resource import APIResource

class Reply(APIResource):

    '''
    Reply Object to be created
    '''
    def __init__(self, endpoint_url=None):
        super().__init__(endpoint_url=endpoint_url)

    @classmethod
    def create(
        cls,
        prompt: str,
        **params,
    ):
            
        '''
        Create a new Reply
        '''

        return cls._post(
                endpoint_url="response/",
                user_prompt=prompt,
                **params,
            )
        
    @classmethod
    def stream(
        cls,
        prompt: str,
        **params,
    ):

        return cls._stream(
                endpoint_url="response_stream/",
                user_prompt=prompt,
                **params,
            )



from prism_ai.api_resources.api_resource import APIResource
from prism_ai.api_resources.knowledge import Knowledge
import pathlib
import os

supported_file_types = [
    "pdf",
    "doc",
    "docx",
    "txt",
    "md",
    "odt",
    "gz"
]

class KnowledgeBase(APIResource):

    '''
    Knowledge Base Object to be created from a url
    '''

    @classmethod
    def add(
        cls, 
        kb_id: int,
        base_dir: str = None,
        base_url: str = None,
        **params,
    ):
        '''
        Add knowledge to an existing knowledge base
        '''

        if not ((base_dir is None) ^ (base_url is None)):
            raise ValueError("Please provide either a base directory or a base url, but not both.")

        if None in [kb_id]:
            raise ValueError("Knowledge Base ID not provided.")

        if not base_url is None: 
            if "name" in params:
                return cls._post(
                    endpoint_url=f"users/knowledge_base/{kb_id}/knowledge_from_url/",
                    base_url=base_url,
                    **params,
                )
            else:
                return cls._post(
                    endpoint_url=f"users/knowledge_base/{kb_id}/knowledge_from_url/",
                    name = "Knowledge pulled from " + base_url,
                    base_url=base_url,
                    **params,
                )
        
        elif not base_dir is None:

            dir_path = pathlib.Path(base_dir)
            dir_list = list(dir_path.rglob('*'))
            file_list = []
            supported_file_list = []
            to_process = []

            for elt in dir_list:

                if elt.is_dir():
                    print(f"Omitting directory {elt} ... Not a file.")
                    continue
                else:
                    file_list.append(elt)

            print("\n\n")
            for elt in file_list:
                
                if str(elt).split(".")[-1] not in supported_file_types:
                    print(f"Omitting file {elt} ... Unsupported file type.")
                    continue
                else:
                    supported_file_list.append(elt)
            print("\n\n")
            for elt in supported_file_list:
                if os.path.getsize(elt) > 1024 * 1024 * 1024 * 4:
                    print(f"Omitting file {elt} ... File size exceeds 4GB.")
                    continue
                else:
                    to_process.append(elt)

            to_process_size = sum([os.path.getsize(file) for file in to_process])
            info_instance = cls._get(endpoint_url="basic_user_info/", quiet=True)
            user_info = info_instance.json

            if user_info["max_storage"] != None:
                if to_process_size / (1024 * 1024) > user_info["max_storage"]:
                    raise ValueError(f"Attempted to upload {to_process_size / (1024 * 1024)} MB of data, but you have only {user_info['max_storage']} MB of storage available. \n\nPlease upgrade your plan, or attempt upload with fewer data.")
            else: 
                pass
            if user_info["tokens_remaining"] <= 0:
                raise ValueError("You have no tokens remaining. Please upgrade your plan to continue using prism.")
            
            for fl in to_process:
                file = pathlib.Path(fl)
                Knowledge.create(
                    method="file",
                    name=file.name,
                    source=file,
                    kb_id=kb_id,
                    **params,
                )
            
            return cls._get(
                endpoint_url=f"users/knowledge_base/{kb_id}/"
            )
        else:
            raise ValueError("Please provide either a base directory or a base url, but not both.")
        
    @classmethod
    def get(
        cls,
        kb_id: int,
        verbosity: bool = False,
        **params,
    ):
            
        '''
        Get a Knowledge Base Object 
        '''

        return cls._get(
                endpoint_url=f"users/knowledge_base/{kb_id}/",
                verbose = verbosity,
            )

    @classmethod
    def create(
        cls,
        name: str,
        **params,
    ):
            
        '''
        Create a new Knowledge Base Object 
        '''

        instance = cls._post(
                endpoint_url=f"users/knowledge_base/",
                name=name,
                **params,
            )
        print(instance.json)
            
        if "base_dir" in params or "base_url" in params:
            instance.add(kb_id=instance.json['id'], **params)

        return instance

    @classmethod
    def delete(
        cls,
        kb_id: int,
        **params,
    ):
            
        '''
        Delete a Knowledge Base Object 
        '''

        return cls._delete(
                endpoint_url=f"users/knowledge_base/{kb_id}/",
                **params,
            )



"""
prism
~~~~~~

The prism python package - a python wrapper for the prism API.
"""

import os

api_key = os.getenv('PRISM_API_KEY')
api_url = "https://api.prism-ai.ch/"
timeout = 30

from prism_ai.api_resources.knowledge import Knowledge
from prism_ai.api_resources.reply import Reply
from prism_ai.api_resources.knowledge_base import KnowledgeBase
from prism_ai.api_resources.api_resource import APIResource


def info(): 

    return APIResource._get(endpoint_url="basic_user_info/")

def __version__():

    """
    Returns information about the current version of the prism package.
    """

    return "0.1.0"


from prism_ai.api_resources.api_resource import APIResource
import requests
from tqdm import tqdm
import os
import pathlib
import httpx
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

supported_file_types = [
    "pdf",
    "doc",
    "docx",
    "txt",
    "md",
    "odt",
    "gz"
]

class Knowledge(APIResource):

    '''
    Knowledge Object to be created
    '''
    
    @classmethod
    def create(
        cls,
        method: str,
        name: str, 
        kb_id: int,
        source: str,
        **params,
    ):
            
        '''
        Create a new Knowledge Object from a url
        '''

        class FileWithProgress:
            def __init__(self, file, total_size, chunk_size=1024*1024):
                self.file = file
                self.total_size = total_size
                self.chunk_size = chunk_size
                self.read_size = 0

            def __iter__(self):
                return self

            def __next__(self):
                data = self.file.read(self.chunk_size)
                if not data:
                    raise StopIteration
                self.read_size += len(data)
                progress_bar.update(len(data))
                return data

        if method == "text":
            return cls._post(
                endpoint_url=f"users/knowledge_base/{kb_id}/knowledge_from_text/",
                name=name,
                text=source,
                **params,
            )

        elif method == "url":
            return cls._post(
                endpoint_url=f"users/knowledge_base/{kb_id}/knowledge_from_url/",
                name=name,
                url=source,
                **params,
            )
            
        elif method == "file":

            try:
                dir_path = pathlib.Path(source)
            except: 
                raise ValueError("The path you provided is not valid.")

            if dir_path.is_dir():

                print("You've provided a directory, to the Knowledge.create method.\n\nPlease use the KnowledgeBase.create method to create a KnowledgeBase from a directory, to create multiple knowledge objects from a directory.")

            elif dir_path.is_file():

                instance = cls(endpoint_url="upload/")
                file_size = os.path.getsize(source)

                info_instance = instance._get(endpoint_url="basic_user_info/", quiet=True)
                user_info = info_instance.json

                if user_info["max_storage"] != None:
                    if file_size / (1024 * 1024) > user_info["max_storage"]:
                        raise ValueError("You have exceeded your storage limit. Please upgrade your plan to continue using prism.")
                else: 
                    pass
                if user_info["tokens_remaining"] <= 0:
                    raise ValueError("You have no tokens remaining. Please upgrade your plan at https://app.prism-ai.ch/ to continue using prism.")
                if file_size > 4 * 1024 * 1024 * 1024:
                    raise ValueError("The file you provided is too large. The maximum file size is 4GB.")
                if str(source).split(".")[-1] not in supported_file_types:
                    raise ValueError("The file you provided is not supported. \nSupported file types are: \n\n - pdf \n - doc \n - docx \n - txt \n - md \n - odt")

                with open(source, 'rb') as file:
                    
                    unique_name = name
                    filename = "kb_"+str(kb_id)+"/"+str(dir_path.name)
                    print("Uploading file "+str(source)+" as "+str(name)+" ...")

                    file_like = FileWithProgress(file, file_size)
                    generate_meta_context = params.pop("generate_meta_context", False)
                    kb_meta_context = params.pop("kb_meta_context", "")

                    headers = instance.create_headers(kb_id=kb_id, unique_name=unique_name, filepath=filename, generate_meta_context=generate_meta_context, kb_meta_context=kb_meta_context)
                    url = instance.api_url + "upload/"

                    with requests.Session() as session: 
                        with tqdm(total=file_size, unit='B', unit_scale=True, dynamic_ncols=True) as progress_bar:
                            response = session.post(url, data=file_like, headers=headers)
                    
                    return response
            else:
                raise ValueError("The path you provided is not valid.")
        else: 
            raise ValueError("The method you provided is not valid. Please use one of the following methods: \n\n - text \n - url \n - file")



from typing import Any
import prism_ai as pai
import requests
import json as jsn
import httpx
from typing import Optional, Any

class APIResource:

    '''
    Parent Class for Prism API Resources 
    '''

    def __init__(
        self,
        endpoint_url: Optional[str],
        response: Optional[requests.Response] = None,
        text: str = "",
        json: dict = {},
        status: int = 0,
    ):

        self.api_key = pai.api_key
        self.api_url = pai.api_url
        self.timeout = pai.timeout
        self.endpoint_url = endpoint_url
        self.response = response
        self.text = text
        self.json = json
        self.status = status

    def __call__(
        self,
        **params: Any,
    ):
        print('This is a parent class for API resource attributes. It is not meant to be called directly. Please refer to our documentation for proper usage of the API.')

    def create_headers(self, kb_id=None, unique_name=None, filepath=None, generate_meta_context=False, kb_meta_context=""):

        headers = {"Authorization":self.api_key}
        if None not in [kb_id, unique_name, filepath]:
            headers['Content-Type'] = 'application/octet-stream'
            headers['Name'] = unique_name
            headers['Filename'] = filepath
            headers['Connection'] = 'keep-alive'
            headers['Keep-Alive'] = '300'
            headers['kb_id'] = str(kb_id)
            headers['generate_meta_context'] = str(generate_meta_context)
            headers['kb_meta_context'] = str(kb_meta_context)
            
        return headers

    def __repr__(self):
        return jsn.dumps(self.json, indent=4, sort_keys=True)

    @classmethod 
    def _delete(
        cls,
        endpoint_url: Optional[str],
        **params,
    ):
            
            '''
            Delete API Resource
            '''
    
            if pai.api_key == None: 
    
                print("Welcome to Prism! To get started, you\'ll need an API Key. \n\n You can get one for free at https://www.prism-ai.ch/\n\nOnce you've got your API key, you can either:\n\n- Set it as an environment variable called PRISM_API_KEY\n- Set it as a variable in your python script called pai.api_key\n\nFor example:\n\nimport prism_ai as pai\npai.api_key = \"YOUR_API_KEY\"\n\nIf you need any help, please refer to our documentation at https://www.prism-ai.ch/docs/ \n\nHappy coding!")
                raise ValueError("API Key is required.")
    
            instance = cls(endpoint_url=endpoint_url)
            json, data, params = instance._prepare_params(**params)
    
            if instance.endpoint_url is None:
                raise ValueError("Endpoint URL is required.")
            else:
                response = requests.delete(
                    instance.api_url + instance.endpoint_url,
                    headers=instance.create_headers(),
                    params=params,
                    data=data,
                    json=json,
                    timeout=instance.timeout,
                )
    
                instance.json = response.json()
                instance.text = response.text
                instance.status = response.status_code
    
                for key, value in instance.json.items():
                    setattr(instance, key, value)
            
            return instance

    @classmethod
    def _get(
        cls,
        endpoint_url: Optional[str],
        quiet = False,
        **params,
    ): 

        '''
        Get API Resource
        '''

        if pai.api_key == None: 

            print("Welcome to Prism! To get started, you\'ll need an API Key. \n\n You can get one for free at https://www.prism-ai.ch/\n\nOnce you've got your API key, you can either:\n\n- Set it as an environment variable called PRISM_API_KEY\n- Set it as a variable in your python script called pai.api_key\n\nFor example:\n\nimport prism_ai as pai\npai.api_key = \"YOUR_API_KEY\"\n\nIf you need any help, please refer to our documentation at https://www.prism-ai.ch/docs/ \n\nHappy coding!")
            raise ValueError("API Key is required.")

        instance = cls(endpoint_url=endpoint_url)
        json, data, params = instance._prepare_params(**params)

        if endpoint_url is None:
            raise ValueError("Endpoint URL is required.")

        else:
            
            if "response_stream" not in endpoint_url:
                response = requests.get(
                        instance.api_url + endpoint_url,
                        headers=instance.create_headers(),
                        params=params,
                        data=data,
                        json=json,
                        timeout=instance.timeout,
                    )

                instance.json = jsn.loads(response.text)
                instance.text = response.text
                instance.status = response.status_code

                for key, value in instance.json.items():
                    setattr(instance, key, value)

        return instance
    

    @classmethod
    def _post(
        cls,
        endpoint_url: Optional[str], 
        **params,
    ):
        '''
        Post API Resource
        '''

        if pai.api_key == None: 

            print("Welcome to Prism! To get started, you\'ll need an API Key. \n\n You can get one for free at https://app.prism-ai.ch/\n\nOnce you've got your API key, you can either:\n\n- Set it as an environment variable called PRISM_API_KEY\n- Set it as a variable in your python script called pai.api_key\n\nFor example:\n\nimport prism_ai as pai\npai.api_key = \"YOUR_API_KEY\"\n\nIf you need any help, please refer to our documentation at https://app.prism-ai.ch/docs/ \n\nHappy coding!")
            
            raise ValueError("API Key is required.")
            
        instance = cls(endpoint_url=endpoint_url)
        json, data, params = instance._prepare_params(**params)

        if instance.endpoint_url is None:
            raise ValueError("Endpoint URL is required.")
        else:
            response = requests.post(
                instance.api_url + instance.endpoint_url,
                headers=instance.create_headers(),
                params=params,
                data=data,
                json=json,
                timeout=instance.timeout,
            )

            instance.json = jsn.loads(response.text)
            instance.text = response.text
            instance.status = response.status_code

            for key, value in instance.json.items():
                setattr(instance, key, value)
        
        return instance

    @classmethod
    def _stream(
        cls, 
        endpoint_url: Optional[str],
        quiet: bool = False,
        **params,
    ):
        
        '''
        Get streaming response from API Resource (outputs an object of type generator)
        '''

        if pai.api_key == None: 

            print("Welcome to Prism! To get started, you\'ll need an API Key. \n\n You can get one for free at https://www.prism-ai.ch/\n\nOnce you've got your API key, you can either:\n\n- Set it as an environment variable called PRISM_API_KEY\n- Set it as a variable in your python script called pai.api_key\n\nFor example:\n\nimport prism_ai as pai\npai.api_key = \"YOUR_API_KEY\"\n\nIf you need any help, please refer to our documentation at https://www.prism-ai.ch/docs/ \n\nHappy coding!")
            raise ValueError("API Key is required.")

        instance = cls(endpoint_url=endpoint_url)
        json, data, params = instance._prepare_params(**params)
        timeout = httpx.Timeout(float(instance.timeout), connect=float(instance.timeout))

        if instance.endpoint_url is None:
            raise ValueError("Endpoint URL is required.")
        
        try:
            
            with httpx.stream(
                method="POST", 
                url=instance.api_url + instance.endpoint_url, 
                headers=instance.create_headers(), 
                timeout = timeout, 
                params=params,
                data=data,
                json=json,
            ) as r:
            
                fulltext = ""
                for chunk in r.iter_text():
                    fulltext += chunk
                    if quiet == False:
                        print(chunk, end="", flush=True)
                    try:
                        yield str(chunk)
                    except KeyError:
                        yield ""
                
        except httpx.RemoteProtocolError as e:

            instance.status = 500
            instance.text = "There\'s been an error on our side. Sorry about that! Our engineers are already contacted and working on a fix. Please try again later."

        except httpx.ConnectTimeout as e:

            instance.status = 500
            instance.text = "There\'s been an error on our side. Sorry about that! Our engineers are already contacted and working on a fix. Please try again later."

    @classmethod
    def _prepare_params(
        cls,
        **params,
    ):
        data = {
            
        }

        json = {
            "name": params.pop("name", None),
            "s3_bucket": params.pop("s3_bucket", None),
            "user_prompt": params.pop("user_prompt", None),
            "knowledge_base": "+".join([str(elt) for elt in params.pop("kb_ids", "")]),
            "text": params.pop("text", None),
            "recursion": params.pop("recursion", False),
            "max_recursion": params.pop("max_recursion", 1),
            "only_base_url": params.pop("only_base_url", True),
            "kb_meta_context": params.pop("kb_meta_context", None),
            "generate_meta_context": params.pop("generate_meta_context", True),
            "smart_index": params.pop("smart_index", False),
            "ner": params.pop("ner", False),
            "verbose": params.pop("verbose", False),
            "model": params.pop("model", "gpt-3.5-turbo-16k"),
            "hyde": params.pop("hyde", False),
            # "prompt": params.pop("prompt", None),
        }
        if "base_url" in params.keys():
            json["url"] = params.pop("base_url", None)
        elif "url" in params.keys():
            json["url"] = params.pop("url", None)
        else:
            pass

        params = {
            "kb_id": params.pop("kb_id", None),
            "knowledge_id": params.pop("knowledge", None),
        }

        return json, data, params
    
